__title__ = "Django REST TypeScript Generator"
__version__ = "0.0.1"
__author__ = "Yinian Chin, Modified by Ukweh Everest"
__license__ = "MIT License"
__copyright__ = "Copyright 2024 Ukweh Chima Everest"

VERSION = __version__
